# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import requests


class HTTPClient(object):
    def __init__(self, base_url):
        self.base_url = base_url
        self.session = self.get_session()

    def get_session(self):
        session = requests.Session()
        session.headers.update(self.default_headers())
        return session

    def default_headers(self):
        return {
            "Content-Type": "application/json",
            "Accept": "application/json",
        }

    def request(self, method, url, **kwargs):
        full_url = self.base_url + url
        resp = self.session.request(method, full_url, **kwargs)
        resp.raise_for_status()
        if resp.headers.get('Content-Type') == 'application/json' and \
                resp.content:
            return resp.json()
        else:
            return None

    def get(self, url, params=None):
        return self.request('GET', url, params=params)

    def put(self, url, body):
        return self.request('PUT', url, json=body)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add server_default to resource_values.values

Revision ID: 967a44dd16d5
Revises: 3b2a0f134e45
Create Date: 2016-03-25 21:12:28.939719

"""

# revision identifiers, used by Alembic.
revision = '967a44dd16d5'
down_revision = '3b2a0f134e45'
branch_labels = None
depends_on = None

from alembic import context
from alembic import op

import tuning_box.db


def upgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    table_name = table_prefix + 'resource_values'
    with op.batch_alter_table(table_name) as batch:
        batch.alter_column(
            'values',
            server_default='{}',
            existing_type=tuning_box.db.Json(),
        )


def downgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    table_name = table_prefix + 'resource_values'
    with op.batch_alter_table(table_name) as batch:
        batch.alter_column(
            'values',
            server_default=None,
            existing_type=tuning_box.db.Json(),
        )



import collections
import decimal
import json
import os
import shutil
import sys

def remove_bs(data):
    new_stdout = []
    for i, e in enumerate(data["stdout"]):
        if '\b' in e[1]:
            c = e[1].count('\b')
            for j, e1 in enumerate(reversed(new_stdout)):
                if len(e1[1]) <= c:
                    c -= len(e1[1])
                    if not c:
                        break
                else:
                    e1[1] = e1[1][:-c]
                    j += 1
                    break
            new_stdout[-j-1:] = []
        else:
            new_stdout.append(e)
    data["stdout"] = new_stdout


def speedup_typing(data, delay):
    delay = decimal.Decimal(delay)
    it = iter(data["stdout"])
    try:
        e = next(it)
        while True:
            while not e[1].endswith("]# "):
                e = next(it)
            e = next(it)
            if e[1] != "#":
                continue
            while True:
                e = next(it)
                e[0] = delay
                if '\r' in e[1]:
                    break
    except StopIteration:
        pass


def set_max_delay(data, delay):
    delay = decimal.Decimal(delay)
    for e in data["stdout"]:
        if e[0] > delay:
            e[0] = delay


def correct_duration(data):
    data["duration"] = sum(e[0] for e in data["stdout"])


def main(fname):
    with open(fname) as f:
        data = json.load(f, object_pairs_hook=collections.OrderedDict, parse_float=decimal.Decimal)

    remove_bs(data)
    speedup_typing(data, '0.05')
    set_max_delay(data, '0.5')
    correct_duration(data)

    shutil.copy(fname, fname + ".tmp")
    with open(fname + ".tmp", "w") as f:
        data = json.dump(data, f, indent=2, cls=Encoder)
    os.rename(fname + ".tmp", fname)


class DecIntWrapper(int):
    def __init__(self, d):
        self.d = d

    def __str__(self):
        return str(self.d)


class Encoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, decimal.Decimal):
            #import pdb; pdb.set_trace()
            return DecIntWrapper(o)
        return json.JSONEncoder.default(o)

if __name__ == '__main__':
    main(sys.argv[1])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from flask import testing
import flask_restful
from werkzeug import exceptions
from werkzeug import wrappers

from tuning_box import app
from tuning_box import db
from tuning_box.tests import base


class JSONResponse(wrappers.BaseResponse):
    @property
    def json(self):
        return json.loads(self.data.decode(self.charset))


class Client(testing.FlaskClient):
    def __init__(self, app):
        super(Client, self).__init__(app, response_wrapper=JSONResponse)

    def open(self, *args, **kwargs):
        data = kwargs.get('data')
        if data is not None:
            kwargs['data'] = json.dumps(data)
            kwargs['content_type'] = 'application/json'
        return super(Client, self).open(*args, **kwargs)


class TestApp(base.TestCase):
    def setUp(self):
        super(TestApp, self).setUp()
        self.app = app.build_app()
        self.app.config["SQLALCHEMY_DATABASE_URI"] = 'sqlite:///'
        with self.app.app_context():
            db.fix_sqlite()
            db.db.create_all()
        self.client = Client(self.app)

    def _fixture(self):
        with self.app.app_context(), db.db.session.begin():
            component = db.Component(
                id=7,
                name='component1',
                resource_definitions=[db.ResourceDefinition(
                    id=5,
                    name='resdef1',
                    content={'key': 'nsname.key'},
                )],
            )
            db.db.session.add(component)
            environment = db.Environment(id=9, components=[component])
            hierarchy_levels = [
                db.EnvironmentHierarchyLevel(name="lvl1"),
                db.EnvironmentHierarchyLevel(name="lvl2"),
            ]
            hierarchy_levels[1].parent = hierarchy_levels[0]
            environment.hierarchy_levels = hierarchy_levels
            db.db.session.add(environment)

    @property
    def _component_json(self):
        return {
            'id': 7,
            'name': 'component1',
            'resource_definitions': [{
                'id': 5,
                'name': 'resdef1',
                'component_id': 7,
                'content': {'key': 'nsname.key'},
            }],
        }

    def _assert_db_effect(self, model, key, fields, expected):
        with self.app.app_context():
            obj = model.query.get(key)
            self.assertIsNotNone(obj)
            marshalled = flask_restful.marshal(obj, fields)
        self.assertEqual(expected, marshalled)

    def _assert_not_in_db(self, model, key):
        with self.app.app_context():
            obj = model.query.get(key)
            self.assertIsNone(obj)

    def test_get_components_empty(self):
        res = self.client.get('/components')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, [])

    def test_get_components(self):
        self._fixture()
        res = self.client.get('/components')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, [self._component_json])

    def test_get_one_component(self):
        self._fixture()
        res = self.client.get('/components/7')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, self._component_json)

    def test_get_one_component_404(self):
        res = self.client.get('/components/7')
        self.assertEqual(res.status_code, 404)

    def test_post_component(self):
        self._fixture()  # Just for namespace
        json = self._component_json
        del json['id']
        del json['resource_definitions'][0]['id']
        del json['resource_definitions'][0]['component_id']
        json['name'] = 'component2'
        res = self.client.post('/components', data=json)
        self.assertEqual(res.status_code, 201)
        json['id'] = 8
        json['resource_definitions'][0]['component_id'] = json['id']
        json['resource_definitions'][0]['id'] = 6
        self.assertEqual(res.json, json)
        self._assert_db_effect(db.Component, 8, app.component_fields, json)

    def test_post_component_conflict(self):
        self._fixture()  # Just for namespace
        json = self._component_json
        del json['id']
        del json['resource_definitions'][0]['id']
        del json['resource_definitions'][0]['component_id']
        res = self.client.post('/components', data=json)
        self.assertEqual(res.status_code, 409)
        self._assert_not_in_db(db.Component, 8)

    def test_post_component_conflict_propagate_exc(self):
        self.app.config["PROPAGATE_EXCEPTIONS"] = True
        self._fixture()  # Just for namespace
        json = self._component_json
        del json['id']
        del json['resource_definitions'][0]['id']
        del json['resource_definitions'][0]['component_id']
        res = self.client.post('/components', data=json)
        self.assertEqual(res.status_code, 409)
        self._assert_not_in_db(db.Component, 8)

    def test_post_component_no_resdef_content(self):
        self._fixture()  # Just for namespace
        json = self._component_json
        del json['id']
        del json['resource_definitions'][0]['id']
        del json['resource_definitions'][0]['component_id']
        del json['resource_definitions'][0]['content']
        json['name'] = 'component2'
        res = self.client.post('/components', data=json)
        self.assertEqual(res.status_code, 201)
        json['id'] = 8
        json['resource_definitions'][0]['component_id'] = json['id']
        json['resource_definitions'][0]['id'] = 6
        json['resource_definitions'][0]['content'] = None
        self.assertEqual(res.json, json)
        self._assert_db_effect(db.Component, 8, app.component_fields, json)

    def test_delete_component(self):
        self._fixture()
        res = self.client.delete('/components/7')
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        self._assert_not_in_db(db.Component, 7)

    def test_delete_component_404(self):
        res = self.client.delete('/components/7')
        self.assertEqual(res.status_code, 404)

    def test_get_environments_empty(self):
        res = self.client.get('/environments')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, [])

    def test_get_environments(self):
        self._fixture()
        res = self.client.get('/environments')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, [{'id': 9, 'components': [7],
                                     'hierarchy_levels': ['lvl1', 'lvl2']}])

    def test_get_one_environment(self):
        self._fixture()
        res = self.client.get('/environments/9')
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, {'id': 9, 'components': [7],
                                    'hierarchy_levels': ['lvl1', 'lvl2']})

    def test_get_one_environment_404(self):
        res = self.client.get('/environments/9')
        self.assertEqual(res.status_code, 404)

    def test_post_environment(self):
        self._fixture()
        json = {'components': [7], 'hierarchy_levels': ['lvla', 'lvlb']}
        res = self.client.post('/environments', data=json)
        self.assertEqual(res.status_code, 201)
        json['id'] = 10
        self.assertEqual(res.json, json)
        self._assert_db_effect(
            db.Environment, 10, app.environment_fields, json)

    def test_post_environment_preserve_id(self):
        self._fixture()
        json = {
            'id': 42,
            'components': [7],
            'hierarchy_levels': ['lvla', 'lvlb'],
        }
        res = self.client.post('/environments', data=json)
        self.assertEqual(res.status_code, 201)
        self.assertEqual(res.json, json)
        self._assert_db_effect(
            db.Environment, 42, app.environment_fields, json)

    def test_post_environment_preserve_id_conflict(self):
        self._fixture()
        json = {
            'id': 9,
            'components': [7],
            'hierarchy_levels': ['lvla', 'lvlb'],
        }
        res = self.client.post('/environments', data=json)
        self.assertEqual(res.status_code, 409)

    def test_post_environment_preserve_id_conflict_propagate_exc(self):
        self.app.config["PROPAGATE_EXCEPTIONS"] = True
        self._fixture()
        json = {
            'id': 9,
            'components': [7],
            'hierarchy_levels': ['lvla', 'lvlb'],
        }
        res = self.client.post('/environments', data=json)
        self.assertEqual(res.status_code, 409)

    def test_post_environment_by_component_name(self):
        self._fixture()
        json = {
            'components': ['component1'],
            'hierarchy_levels': ['lvla', 'lvlb'],
        }
        res = self.client.post('/environments', data=json)
        self.assertEqual(res.status_code, 201)
        json['id'] = 10
        json['components'] = [7]
        self.assertEqual(res.json, json)
        self._assert_db_effect(
            db.Environment, 10, app.environment_fields, json)

    def test_post_environment_404(self):
        self._fixture()
        json = {'components': [8], 'hierarchy_levels': ['lvla', 'lvlb']}
        res = self.client.post('/environments', data=json)
        self.assertEqual(res.status_code, 404)
        self._assert_not_in_db(db.Environment, 10)

    def test_post_environment_by_component_name_404(self):
        self._fixture()
        json = {
            'components': ['component2'],
            'hierarchy_levels': ['lvla', 'lvlb'],
        }
        res = self.client.post('/environments', data=json)
        self.assertEqual(res.status_code, 404)
        self._assert_not_in_db(db.Environment, 10)

    def test_delete_environment(self):
        self._fixture()
        res = self.client.delete('/environments/9')
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        self._assert_not_in_db(db.Environment, 9)

    def test_delete_environment_404(self):
        res = self.client.delete('/environments/9')
        self.assertEqual(res.status_code, 404)

    def test_get_environment_level_value_root(self):
        self._fixture()
        with self.app.app_context(), db.db.session.begin():
            level_value = app.get_environment_level_value(
                db.Environment(id=9),
                [],
            )
            self.assertIsNone(level_value)

    def test_get_environment_level_value_deep(self):
        self._fixture()
        with self.app.app_context(), db.db.session.begin():
            level_value = app.get_environment_level_value(
                db.Environment(id=9),
                [('lvl1', 'val1'), ('lvl2', 'val2')],
            )
            self.assertIsNotNone(level_value)
            self.assertEqual(level_value.level.name, 'lvl2')
            self.assertEqual(level_value.value, 'val2')
            level_value = level_value.parent
            self.assertEqual(level_value.level.name, 'lvl1')
            self.assertEqual(level_value.value, 'val1')
            self.assertIsNone(level_value.parent)

    def test_get_environment_level_value_bad_level(self):
        self._fixture()
        with self.app.app_context(), db.db.session.begin():
            exc = self.assertRaises(
                exceptions.BadRequest,
                app.get_environment_level_value,
                db.Environment(id=9),
                [('lvlx', 'val1')],
            )
            self.assertEqual(
                exc.description,
                "Unexpected level name 'lvlx'. Expected 'lvl1'.",
            )

    def test_put_resource_values_root(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/values',
                              data={'k': 'v'})
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        with self.app.app_context():
            resource_values = db.ResourceValues.query.filter_by(
                environment_id=9, resource_definition_id=5).one_or_none()
            self.assertIsNotNone(resource_values)
            self.assertEqual(resource_values.values, {'k': 'v'})
            self.assertIsNone(resource_values.level_value)

    def test_put_resource_values_deep(self):
        self._fixture()
        res = self.client.put(
            '/environments/9/lvl1/val1/lvl2/val2/resources/5/values',
            data={'k': 'v'},
        )
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        with self.app.app_context():
            resource_values = db.ResourceValues.query.filter_by(
                environment_id=9, resource_definition_id=5).one_or_none()
            self.assertIsNotNone(resource_values)
            self.assertEqual(resource_values.values, {'k': 'v'})
            level_value = resource_values.level_value
            self.assertEqual(level_value.level.name, 'lvl2')
            self.assertEqual(level_value.value, 'val2')
            level_value = level_value.parent
            self.assertEqual(level_value.level.name, 'lvl1')
            self.assertEqual(level_value.value, 'val1')
            self.assertIsNone(level_value.parent)

    def test_put_resource_values_overrides_root(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/overrides',
                              data={'k': 'v'})
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        with self.app.app_context():
            resource_values = db.ResourceValues.query.filter_by(
                environment_id=9, resource_definition_id=5).one_or_none()
            self.assertIsNotNone(resource_values)
            self.assertEqual(resource_values.overrides, {'k': 'v'})
            self.assertIsNone(resource_values.level_value)

    def test_put_resource_values_overrides_deep(self):
        self._fixture()
        res = self.client.put(
            '/environments/9/lvl1/val1/lvl2/val2/resources/5/overrides',
            data={'k': 'v'},
        )
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        with self.app.app_context():
            resource_values = db.ResourceValues.query.filter_by(
                environment_id=9, resource_definition_id=5).one_or_none()
            self.assertIsNotNone(resource_values)
            self.assertEqual(resource_values.overrides, {'k': 'v'})
            level_value = resource_values.level_value
            self.assertEqual(level_value.level.name, 'lvl2')
            self.assertEqual(level_value.value, 'val2')
            level_value = level_value.parent
            self.assertEqual(level_value.level.name, 'lvl1')
            self.assertEqual(level_value.value, 'val1')
            self.assertIsNone(level_value.parent)

    def test_put_resource_values_bad_level(self):
        self._fixture()
        res = self.client.put('/environments/9/lvlx/1/resources/5/values',
                              data={'k': 'v'})
        self.assertEqual(res.status_code, 400)
        self.assertEqual(
            res.json,
            {"message": "Unexpected level name 'lvlx'. Expected 'lvl1'."},
        )
        with self.app.app_context():
            resource_values = db.ResourceValues.query.filter_by(
                environment_id=9, resource_definition_id=5).one_or_none()
            self.assertIsNone(resource_values)

    def test_get_resource_values(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/values',
                              data={'key': 'value'})
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        res = self.client.get(
            '/environments/9/lvl1/1/resources/5/values',
        )
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, {})

    def test_get_resource_values_effective(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/values',
                              data={'key': 'value'})
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        res = self.client.get(
            '/environments/9/lvl1/1/resources/5/values?effective',
        )
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, {'key': 'value'})

    def test_get_resource_values_local_override(self):
        self._fixture()
        res = self.client.put('/environments/9/lvl1/1/resources/5/values',
                              data={'key': 'value1'})
        res = self.client.put('/environments/9/lvl1/1/resources/5/overrides',
                              data={'key': 'value2'})
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        res = self.client.get(
            '/environments/9/lvl1/1/resources/5/values?effective',
        )
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, {'key': 'value2'})

    def test_get_resource_values_level_override(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/values',
                              data={'key': 'value', 'key1': 'value'})
        res = self.client.put('/environments/9/lvl1/1/resources/5/values',
                              data={'key': 'value1'})
        res = self.client.put('/environments/9/lvl1/2/resources/5/values',
                              data={'key1': 'value2'})
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        res = self.client.get(
            '/environments/9/lvl1/1/resources/5/values?effective',
        )
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, {'key': 'value1', 'key1': 'value'})

    def test_get_resource_values_level_and_local_override(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/values',
                              data={'key': 'value', 'key1': 'value'})
        res = self.client.put('/environments/9/lvl1/1/resources/5/values',
                              data={'key': 'value1'})
        res = self.client.put('/environments/9/lvl1/1/resources/5/overrides',
                              data={'key1': 'value2'})
        self.assertEqual(res.status_code, 204)
        self.assertEqual(res.data, b'')
        res = self.client.get(
            '/environments/9/lvl1/1/resources/5/values?effective',
        )
        self.assertEqual(res.status_code, 200)
        self.assertEqual(res.json, {'key': 'value1', 'key1': 'value2'})

    def test_put_resource_values_redirect(self):
        self._fixture()
        res = self.client.put(
            '/environments/9/lvl1/val1/lvl2/val2/resources/resdef1/values',
            data={'k': 'v'},
        )
        self.assertEqual(res.status_code, 308)
        self.assertEqual(
            res.headers['Location'],
            'http://localhost'
            '/environments/9/lvl1/val1/lvl2/val2/resources/5/values',
        )

    def test_get_resource_values_redirect(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/values',
                              data={'key': 'value'})
        res = self.client.get(
            '/environments/9/lvl1/val1/lvl2/val2/resources/resdef1/values',
        )
        self.assertEqual(res.status_code, 308)
        self.assertEqual(
            res.headers['Location'],
            'http://localhost'
            '/environments/9/lvl1/val1/lvl2/val2/resources/5/values',
        )

    def test_get_resource_values_redirect_with_query(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/values',
                              data={'key': 'value'})
        res = self.client.get(
            '/environments/9/lvl1/val1/lvl2/val2/resources/resdef1/values'
            '?effective',
        )
        self.assertEqual(res.status_code, 308)
        self.assertEqual(
            res.headers['Location'],
            'http://localhost'
            '/environments/9/lvl1/val1/lvl2/val2/resources/5/values?effective',
        )

    def test_put_resource_overrides_redirect(self):
        self._fixture()
        res = self.client.put(
            '/environments/9/lvl1/val1/lvl2/val2/resources/resdef1/overrides',
            data={'k': 'v'},
        )
        self.assertEqual(res.status_code, 308)
        self.assertEqual(
            res.headers['Location'],
            'http://localhost'
            '/environments/9/lvl1/val1/lvl2/val2/resources/5/overrides',
        )

    def test_get_resource_overrides_redirect(self):
        self._fixture()
        res = self.client.put('/environments/9/resources/5/overrides',
                              data={'key': 'value'})
        res = self.client.get(
            '/environments/9/lvl1/val1/lvl2/val2/resources/resdef1/overrides',
        )
        self.assertEqual(res.status_code, 308)
        self.assertEqual(
            res.headers['Location'],
            'http://localhost'
            '/environments/9/lvl1/val1/lvl2/val2/resources/5/overrides',
        )


class TestAppPrefixed(base.PrefixedTestCaseMixin, TestApp):
    pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import itertools

import flask
import flask_restful
from flask_restful import fields
from sqlalchemy import exc as sa_exc
from werkzeug import exceptions

from tuning_box import converters
from tuning_box import db

# These handlers work if PROPAGATE_EXCEPTIONS is off (non-Nailgun case)
api_errors = {
    'IntegrityError': {'status': 409},  # sqlalchemy IntegrityError
}
api = flask_restful.Api(errors=api_errors)

resource_definition_fields = {
    'id': fields.Integer,
    'name': fields.String,
    'component_id': fields.Integer,
    'content': fields.Raw,
}

component_fields = {
    'id': fields.Integer,
    'name': fields.String,
    'resource_definitions': fields.List(
        fields.Nested(resource_definition_fields)),
}


def with_transaction(f):
    @functools.wraps(f)
    def inner(*args, **kwargs):
        with db.db.session.begin():
            return f(*args, **kwargs)

    return inner


@api.resource('/components')
class ComponentsCollection(flask_restful.Resource):
    method_decorators = [flask_restful.marshal_with(component_fields)]

    def get(self):
        return db.Component.query.all()

    @with_transaction
    def post(self):
        component = db.Component(name=flask.request.json['name'])
        component.resource_definitions = []
        for resdef_data in flask.request.json.get('resource_definitions'):
            resdef = db.ResourceDefinition(name=resdef_data['name'],
                                           content=resdef_data.get('content'))
            component.resource_definitions.append(resdef)
        db.db.session.add(component)
        return component, 201


@api.resource('/components/<int:component_id>')
class Component(flask_restful.Resource):
    method_decorators = [flask_restful.marshal_with(component_fields)]

    def get(self, component_id):
        return db.Component.query.get_or_404(component_id)

    @with_transaction
    def delete(self, component_id):
        component = db.Component.query.get_or_404(component_id)
        db.db.session.delete(component)
        return None, 204

environment_fields = {
    'id': fields.Integer,
    'components': fields.List(fields.Integer(attribute='id')),
    'hierarchy_levels': fields.List(fields.String(attribute='name')),
}


@api.resource('/environments')
class EnvironmentsCollection(flask_restful.Resource):
    method_decorators = [flask_restful.marshal_with(environment_fields)]

    def get(self):
        return db.Environment.query.all()

    @with_transaction
    def post(self):
        component_ids = flask.request.json['components']
        # TODO(yorik-sar): verify that resource names do not clash
        components = [db.Component.query.get_by_id_or_name(i)
                      for i in component_ids]

        hierarchy_levels = []
        level = None
        for name in flask.request.json['hierarchy_levels']:
            level = db.EnvironmentHierarchyLevel(name=name, parent=level)
            hierarchy_levels.append(level)

        environment = db.Environment(components=components,
                                     hierarchy_levels=hierarchy_levels)
        if 'id' in flask.request.json:
            environment.id = flask.request.json['id']
        db.db.session.add(environment)
        return environment, 201


@api.resource('/environments/<int:environment_id>')
class Environment(flask_restful.Resource):
    method_decorators = [flask_restful.marshal_with(environment_fields)]

    def get(self, environment_id):
        return db.Environment.query.get_or_404(environment_id)

    @with_transaction
    def delete(self, environment_id):
        environment = db.Environment.query.get_or_404(environment_id)
        db.db.session.delete(environment)
        return None, 204


def iter_environment_level_values(environment, levels):
    env_levels = db.EnvironmentHierarchyLevel.get_for_environment(environment)
    level_pairs = zip(env_levels, levels)
    parent_level_value = None
    for env_level, (level_name, level_value) in level_pairs:
        if env_level.name != level_name:
            raise exceptions.BadRequest(
                "Unexpected level name '%s'. Expected '%s'." % (
                    level_name, env_level.name))
        level_value_db = db.get_or_create(
            db.EnvironmentHierarchyLevelValue,
            level=env_level,
            parent=parent_level_value,
            value=level_value,
        )
        yield level_value_db
        parent_level_value = level_value_db


def get_environment_level_value(environment, levels):
    level_value = None
    for level_value in iter_environment_level_values(environment, levels):
        pass
    return level_value


@api.resource(
    '/environments/<int:environment_id>' +
    '/<levels:levels>resources/<id_or_name:resource_id_or_name>/values')
class ResourceValues(flask_restful.Resource):
    @with_transaction
    def put(self, environment_id, levels, resource_id_or_name):
        environment = db.Environment.query.get_or_404(environment_id)
        level_value = get_environment_level_value(environment, levels)
        # TODO(yorik-sar): filter by environment
        resdef = db.ResourceDefinition.query.get_by_id_or_name(
            resource_id_or_name)
        if resdef.id != resource_id_or_name:
            return flask.redirect(api.url_for(
                ResourceValues,
                environment_id=environment_id,
                levels=levels,
                resource_id_or_name=resdef.id,
            ), code=308)
        esv = db.get_or_create(
            db.ResourceValues,
            environment=environment,
            resource_definition=resdef,
            level_value=level_value,
        )
        esv.values = flask.request.json
        return None, 204

    @with_transaction
    def get(self, environment_id, resource_id_or_name, levels):
        environment = db.Environment.query.get_or_404(environment_id)
        level_values = list(iter_environment_level_values(environment, levels))
        # TODO(yorik-sar): filter by environment
        resdef = db.ResourceDefinition.query.get_by_id_or_name(
            resource_id_or_name)
        if resdef.id != resource_id_or_name:
            url = api.url_for(
                ResourceValues,
                environment_id=environment_id,
                levels=levels,
                resource_id_or_name=resdef.id,
            )
            if flask.request.query_string:
                qs = flask.request.query_string.decode('utf-8')
                url += '?' + qs
            return flask.redirect(url, code=308)
        if 'effective' in flask.request.args:
            resource_values = db.ResourceValues.query.filter_by(
                resource_definition=resdef,
                environment=environment,
            ).all()
            result = {}
            for level_value in itertools.chain([None], level_values):
                for resource_value in resource_values:
                    if resource_value.level_value == level_value:
                        result.update(resource_value.values)
                        result.update(resource_value.overrides)
                        break
            return result
        else:
            if not level_values:
                level_value = None
            else:
                level_value = level_values[-1]
            resource_values = db.ResourceValues.query.filter_by(
                resource_definition=resdef,
                environment=environment,
                level_value=level_value,
            ).one_or_none()
            if not resource_values:
                return {}
            return resource_values.values


@api.resource(
    '/environments/<int:environment_id>' +
    '/<levels:levels>resources/<id_or_name:resource_id_or_name>/overrides')
class ResourceOverrides(flask_restful.Resource):
    @with_transaction
    def put(self, environment_id, levels, resource_id_or_name):
        environment = db.Environment.query.get_or_404(environment_id)
        level_value = get_environment_level_value(environment, levels)
        # TODO(yorik-sar): filter by environment
        resdef = db.ResourceDefinition.query.get_by_id_or_name(
            resource_id_or_name)
        if resdef.id != resource_id_or_name:
            return flask.redirect(api.url_for(
                ResourceOverrides,
                environment_id=environment_id,
                levels=levels,
                resource_id_or_name=resdef.id,
            ), code=308)
        esv = db.get_or_create(
            db.ResourceValues,
            environment=environment,
            resource_definition=resdef,
            level_value=level_value,
        )
        esv.overrides = flask.request.json
        return None, 204

    @with_transaction
    def get(self, environment_id, resource_id_or_name, levels):
        environment = db.Environment.query.get_or_404(environment_id)
        level_value = get_environment_level_value(environment, levels)
        # TODO(yorik-sar): filter by environment
        resdef = db.ResourceDefinition.query.get_by_id_or_name(
            resource_id_or_name)
        if resdef.id != resource_id_or_name:
            url = api.url_for(
                ResourceOverrides,
                environment_id=environment_id,
                levels=levels,
                resource_id_or_name=resdef.id,
            )
            return flask.redirect(url, code=308)
        resource_values = db.ResourceValues.query.filter_by(
            resource_definition=resdef,
            environment=environment,
            level_value=level_value,
        ).one_or_none()
        if not resource_values:
            return {}
        return resource_values.overrides


def handle_integrity_error(exc):
    response = flask.jsonify(msg=exc.args[0])
    response.status_code = 409
    return response


def build_app():
    app = flask.Flask(__name__)
    app.url_map.converters.update(converters.ALL)
    api.init_app(app)  # init_app spoils Api object if app is a blueprint
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False  # silence warning
    # These handlers work if PROPAGATE_EXCEPTIONS is on (Nailgun case)
    app.register_error_handler(sa_exc.IntegrityError, handle_integrity_error)
    db.db.init_app(app)
    return app


def main():
    import logging
    logging.basicConfig(level=logging.DEBUG)

    app = build_app()
    app.run()

if __name__ == '__main__':
    main()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add unique constraint on component.name

Revision ID: d054eefc4c5b
Revises: ad192a40fd68
Create Date: 2016-03-25 21:34:54.487361

"""

# revision identifiers, used by Alembic.
revision = 'd054eefc4c5b'
down_revision = 'ad192a40fd68'
branch_labels = None
depends_on = None

from alembic import context
from alembic import op


def upgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    table_name = table_prefix + 'component'
    with op.batch_alter_table(table_name) as batch:
        batch.create_unique_constraint(
            table_name + '_component_name_unique',
            ['name'],
        )


def downgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    table_name = table_prefix + 'component'
    with op.batch_alter_table(table_name) as batch:
        batch.drop_constraint(
            table_name + '_component_name_unique',
            type_='unique',
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

import itertools
import threading

from nailgun.db import sqlalchemy as nailgun_sa
from nailgun import extensions
import web

import tuning_box
from tuning_box import app as tb_app
from tuning_box import db as tb_db


class App2WebPy(web.application):
    def __init__(self):
        web.application.__init__(self)
        self.__name__ = self
        self.app = None
        self.lock = threading.Lock()

    def create_app(self):
        raise NotImplementedError

    def get_app(self):
        with self.lock:
            if not self.app:
                self.app = self.create_app()
            return self.app

    def handle(self):
        written_data = []

        def write(data):
            assert start_response.called
            written_data.append(data)

        def start_response(status, headers, exc_info=None):
            assert not start_response.called
            assert not exc_info
            start_response.called = True
            web.ctx.status = status
            web.ctx.headers.extend(headers)
            return write

        start_response.called = False

        app = self.get_app()
        environ = dict(web.ctx.environ)
        environ["SCRIPT_NAME"] = environ["PATH_INFO"][:-len(web.ctx.path)]
        environ["PATH_INFO"] = environ["REQUEST_URI"] = web.ctx.path
        result = app(environ, start_response)
        return itertools.chain(written_data, result)


class TB2WebPy(App2WebPy):
    def create_app(self):
        app = tb_app.build_app()
        tb_db.prefix_tables(tb_db, Extension.table_prefix())
        app.config["PROPAGATE_EXCEPTIONS"] = True
        app.config["SQLALCHEMY_DATABASE_URI"] = nailgun_sa.db_str
        return app


class Extension(extensions.BaseExtension):
    name = 'tuning_box'
    version = tuning_box.__version__
    description = 'Plug tuning_box endpoints into Nailgun itself'

    urls = [{'uri': '/config', 'handler': TB2WebPy()}]

    @classmethod
    def alembic_migrations_path(cls):
        return tuning_box.get_migrations_dir()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Remove fake root hierarchy level values

Revision ID: 9ae15c85fa92
Revises: d054eefc4c5b
Create Date: 2016-04-12 20:22:06.323291

"""

# revision identifiers, used by Alembic.
revision = '9ae15c85fa92'
down_revision = 'd054eefc4c5b'
branch_labels = None
depends_on = None

from alembic import context
from alembic import op
import sqlalchemy as sa
from sqlalchemy.ext import automap


def _get_autobase(table_prefix, bind):
    metadata = sa.MetaData(bind=bind)
    table_name = table_prefix + 'environment_hierarchy_level_value'
    metadata.reflect(only=[table_name])
    AutoBase = automap.automap_base(metadata=metadata)

    def classname_for_table(base, refl_table_name, table):
        assert refl_table_name.startswith(table_prefix)
        noprefix_name = refl_table_name[len(table_prefix):]
        uname = u"".join(s.capitalize() for s in noprefix_name.split('_'))
        if not isinstance(uname, str):
            return uname.encode('utf-8')
        else:
            return uname

    AutoBase.prepare(classname_for_table=classname_for_table)
    return AutoBase


def _get_ehlv_class():
    table_prefix = context.config.get_main_option('table_prefix')
    bind = op.get_bind()
    AutoBase = _get_autobase(table_prefix, bind)
    return AutoBase.classes.EnvironmentHierarchyLevelValue


def _get_session():
    return sa.orm.Session(bind=op.get_bind(), autocommit=True)


def upgrade():
    EHLV = _get_ehlv_class()
    session = _get_session()
    with session.begin():
        fake_roots = session.query(EHLV) \
            .filter_by(level_id=None, parent_id=None, value=None) \
            .all()
        if fake_roots:
            fake_root_ids = [r.id for r in fake_roots]
            session.query(EHLV) \
                .filter(EHLV.parent_id.in_(fake_root_ids)) \
                .update({EHLV.parent_id: None}, synchronize_session=False)
            for r in fake_roots:
                session.delete(r)


def downgrade():
    EHLV = _get_ehlv_class()
    session = _get_session()
    with session.begin():
        fake_root = EHLV(level_id=None, parent_id=None, value=None)
        session.add(fake_root)
        session.flush()
        session.query(EHLV) \
            .filter(EHLV.parent_id == None,  # noqa
                    EHLV.id != fake_root.id) \
            .update({EHLV.parent_id: fake_root.id},
                    synchronize_session=False)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Switch to new API

Revision ID: 3b2a0f134e45
Revises: f16eb4eff7c
Create Date: 2016-03-17 16:30:11.989340

"""

# revision identifiers, used by Alembic.
revision = '3b2a0f134e45'
down_revision = 'f16eb4eff7c'
branch_labels = None
depends_on = None

from alembic import context
from alembic import op
import sqlalchemy as sa

import tuning_box.db


def upgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    op.drop_table(table_prefix + 'template')
    table_name = table_prefix + 'environment_schema_values'
    with op.batch_alter_table(table_name) as batch:
        batch.drop_constraint(table_name + '_schema_id_fkey', 'foreignkey')
        batch.alter_column(
            'schema_id',
            new_column_name='resource_definition_id',
            existing_type=sa.Integer(),
        )
    op.rename_table(table_name, table_prefix + 'resource_values')
    op.rename_table(table_prefix + 'schema',
                    table_prefix + 'resource_definition')
    with op.batch_alter_table(table_prefix + 'resource_definition') as batch:
        batch.drop_column('namespace_id')
    op.drop_table(table_prefix + 'namespace')
    table_name = table_prefix + 'resource_values'
    with op.batch_alter_table(table_name) as batch:
        batch.create_foreign_key(
            table_name + '_resource_definition_id_fkey',
            table_prefix + 'resource_definition',
            ['resource_definition_id'],
            ['id'],
        )


def downgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    table_name = table_prefix + 'resource_values'
    with op.batch_alter_table(table_name) as batch:
        batch.drop_constraint(table_name + '_resource_definition_id_fkey',
                              'foreignkey')
    op.create_table(
        table_prefix + 'namespace',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('name', sa.String(length=128), nullable=True),
    )
    table_name = table_prefix + 'schema'
    op.rename_table(table_prefix + 'resource_definition', table_name)
    with op.batch_alter_table(table_name) as batch:
        batch.add_column(
            sa.Column('namespace_id', sa.Integer(), nullable=True))
    table_name = table_prefix + 'environment_schema_values'
    op.rename_table(table_prefix + 'resource_values', table_name)
    with op.batch_alter_table(table_name) as batch:
        batch.alter_column(
            'resource_definition_id',
            new_column_name='schema_id',
            existing_type=sa.Integer(),
        )
        batch.create_foreign_key(
            table_name + '_schema_id_fkey',
            table_prefix + 'schema',
            ['schema_id'],
            ['id'],
        )
    table_name = table_prefix + 'template'
    op.create_table(
        table_name,
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('name', sa.String(length=128), nullable=True),
        sa.Column('component_id', sa.Integer(), nullable=True),
        sa.Column('content', tuning_box.db.Json(), nullable=True),
        sa.ForeignKeyConstraint(
            ['component_id'], [table_prefix + 'component.id'],
            name=table_name + '_component_id_fkey',
        ),
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import absolute_import

from cliff import command
from fuelclient import client as fc_client

from tuning_box import cli
from tuning_box.cli import base as cli_base
from tuning_box.cli import get as cli_get
from tuning_box.cli import set as cli_set
from tuning_box import client as tb_client


class FuelHTTPClient(tb_client.HTTPClient):
    def __init__(self):
        service_catalog = fc_client.APIClient.keystone_client.service_catalog
        base_url = service_catalog.url_for(
            service_type='config',
            endpoint_type='publicURL',
        )
        super(FuelHTTPClient, self).__init__(base_url)

    def default_headers(self):
        headers = super(FuelHTTPClient, self).default_headers()
        if fc_client.APIClient.auth_token:
            headers['X-Auth-Token'] = fc_client.APIClient.auth_token
        return headers


class FuelBaseCommand(cli_base.BaseCommand):
    def get_client(self):
        return FuelHTTPClient()


class Get(FuelBaseCommand, cli_get.Get):
    pass


class Set(FuelBaseCommand, cli_set.Set):
    pass


class Override(FuelBaseCommand, cli_set.Override):
    pass


class Config(command.Command):
    def get_parser(self, *args, **kwargs):
        parser = super(Config, self).get_parser(*args, **kwargs)
        parser.add_argument('args', nargs='*')
        return parser

    def take_action(self, parsed_args):
        client = FuelHTTPClient()
        app = cli.TuningBoxApp(client)
        app.run(parsed_args.args)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import functools
import json
import re

import flask
import flask_sqlalchemy
import sqlalchemy.event
import sqlalchemy.ext.declarative as sa_decl
from sqlalchemy.orm import exc as orm_exc
from sqlalchemy import types

try:
    from importlib import reload
except ImportError:
    pass  # in 2.x reload is builtin

db = flask_sqlalchemy.SQLAlchemy(session_options={'autocommit': True})
pk_type = db.Integer
pk = functools.partial(db.Column, pk_type, primary_key=True)


def fk(cls, **kwargs):
    return db.Column(pk_type, db.ForeignKey(cls.id), **kwargs)


class BaseQuery(flask_sqlalchemy.BaseQuery):
    def get_by_id_or_name(self, id_or_name):
        if isinstance(id_or_name, int):
            result = self.get(id_or_name)
        else:
            result = self.filter_by(name=id_or_name).one_or_none()
        if result is None:
            flask.abort(404)
        return result

    # one_or_none is not present in sqlalchemy < 1.0.9
    def one_or_none(self):
        ret = list(self)
        l = len(ret)
        if l == 1:
            return ret[0]
        elif l == 0:
            return None
        else:
            raise orm_exc.MultipleResultsFound(
                "Multiple rows were found for one_or_none()")


def _tablename(cls_name):
    def repl(match):
        res = match.group().lower()
        if match.start():
            res = "_" + res
        return res

    return ModelMixin.table_prefix + re.sub("[A-Z]", repl, cls_name)


class ModelMixin(object):
    query_class = BaseQuery
    id = db.Column(pk_type, primary_key=True)

    try:
        table_prefix = ModelMixin.table_prefix  # keep prefix during reload
    except NameError:
        table_prefix = ""  # first import, not reload

    @sa_decl.declared_attr
    def __tablename__(cls):
        return _tablename(cls.__name__)

    def __repr__(self):
        args = []
        for attr in self.__repr_attrs__:
            value = getattr(self, attr)
            if attr == 'content' and value is not None and len(value) > 15:
                value = value[:10] + '<...>'
            args.append('{}={!r}'.format(attr, value))
        return '{}({})'.format(type(self).__name__, ','.join(args))


class Json(types.TypeDecorator):
    impl = db.Text

    def process_bind_param(self, value, dialect):
        return json.dumps(value)

    def process_result_value(self, value, dialect):
        return json.loads(value)


# Component registry


class Component(ModelMixin, db.Model):
    name = db.Column(db.String(128), unique=True)

    __repr_attrs__ = ('id', 'name')


class ResourceDefinition(ModelMixin, db.Model):
    name = db.Column(db.String(128))
    component_id = fk(Component)
    component = db.relationship(Component, backref='resource_definitions')
    content = db.Column(Json)

    __repr_attrs__ = ('id', 'name', 'component', 'content')

# Environment data storage


class Environment(ModelMixin, db.Model):
    @sa_decl.declared_attr
    def environment_components_table(cls):
        return db.Table(
            _tablename('environment_components'),
            db.Column('environment_id', pk_type, db.ForeignKey(cls.id)),
            db.Column('component_id', pk_type, db.ForeignKey(Component.id)),
        )

    @sa_decl.declared_attr
    def components(cls):
        return db.relationship(
            Component, secondary=cls.environment_components_table)

    __repr_attrs__ = ('id',)


class EnvironmentHierarchyLevel(ModelMixin, db.Model):
    environment_id = fk(Environment)
    environment = db.relationship(Environment, backref='hierarchy_levels')
    name = db.Column(db.String(128))

    @sa_decl.declared_attr
    def parent_id(cls):
        return db.Column(pk_type, db.ForeignKey(cls.id))

    @sa_decl.declared_attr
    def parent(cls):
        return db.relationship(cls,
                               backref=db.backref('child', uselist=False),
                               remote_side=cls.id)

    __table_args__ = (
        db.UniqueConstraint('environment_id', 'name'),
        db.UniqueConstraint('environment_id', 'parent_id'),
    )
    __repr_attrs__ = ('id', 'environment', 'parent', 'name')

    @classmethod
    def get_for_environment(cls, environment):
        query = cls.query.filter_by(environment=environment, parent=None)
        root_level = query.one_or_none()
        if not root_level:
            return []
        env_levels = [root_level]
        while env_levels[-1].child:
            env_levels.append(env_levels[-1].child)
        return env_levels


class EnvironmentHierarchyLevelValue(ModelMixin, db.Model):
    level_id = fk(EnvironmentHierarchyLevel)
    level = db.relationship(EnvironmentHierarchyLevel)
    value = db.Column(db.String(128))

    @sa_decl.declared_attr
    def parent_id(cls):
        return db.Column(pk_type, db.ForeignKey(cls.id))

    @sa_decl.declared_attr
    def parent(cls):
        return db.relationship(cls, remote_side=cls.id)

    # TODO(yorik-sar): add UniqueConstraint for all fields
    __repr_attrs__ = ('id', 'level', 'parent', 'value')


class ResourceValues(ModelMixin, db.Model):
    environment_id = fk(Environment)
    environment = db.relationship(Environment)
    resource_definition_id = fk(ResourceDefinition)
    resource_definition = db.relationship(ResourceDefinition)
    level_value_id = fk(EnvironmentHierarchyLevelValue)
    level_value = db.relationship('EnvironmentHierarchyLevelValue')
    values = db.Column(Json, server_default='{}')
    overrides = db.Column(Json, server_default='{}')

    __table_args__ = (
        db.UniqueConstraint(environment_id, resource_definition_id,
                            level_value_id),
    )
    __repr_attrs__ = ('id', 'environment', 'resource_definition',
                      'level_value', 'values')


def get_or_create(cls, **attrs):
    with db.session.begin(nested=True):
        item = cls.query.filter_by(**attrs).one_or_none()
        if not item:
            item = cls(**attrs)
            db.session.add(item)
            # TODO(yorik-sar): handle constraints failure in case of
            # race condition
    return item


def fix_sqlite():
    engine = db.engine

    @sqlalchemy.event.listens_for(engine, "connect")
    def _connect(dbapi_connection, connection_record):
        dbapi_connection.isolation_level = None

    @sqlalchemy.event.listens_for(engine, "begin")
    def _begin(conn):
        conn.execute("BEGIN")


def prefix_tables(module, prefix):
    ModelMixin.table_prefix = prefix
    reload(module)


def unprefix_tables(module):
    ModelMixin.table_prefix = ""
    reload(module)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tuning_box.cli import base


class Get(base.FormattedCommand, base.ResourceCommand):
    def get_parser(self, *args, **kwargs):
        parser = super(Get, self).get_parser(*args, **kwargs)
        parser.add_argument(
            '--key',
            type=str,
            help="Name of key to get from the resource",
        )
        return parser

    def take_action(self, parsed_args):
        res = self.get_client().get(
            self.get_resource_url(parsed_args),
            params='effective',
        )
        key = parsed_args.key
        if key is None:
            return res
        value = res[key]
        if parsed_args.format != 'plain':
            return {key: value}
        else:
            return value



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os

# oslo_db internals refuse to work properly if this is not set
# actual file name in that URL doesn't matter, it'll be generated by oslo.db
os.environ.setdefault("OS_TEST_DBAPI_ADMIN_CONNECTION", "sqlite:///testdb")

from alembic import command as alembic_command
from alembic import config as alembic_config
from alembic import script as alembic_script
import flask
from oslo_db.sqlalchemy import test_base
from oslo_db.sqlalchemy import test_migrations
import sqlalchemy as sa
import testscenarios
from werkzeug import exceptions

import tuning_box
from tuning_box import db
from tuning_box.tests import base


class _DBTestCase(base.TestCase):
    def setUp(self):
        super(_DBTestCase, self).setUp()
        self.app = flask.Flask('test')
        self.app.config["SQLALCHEMY_DATABASE_URI"] = 'sqlite:///'
        self.app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False  # no warning
        db.db.init_app(self.app)
        with self.app.app_context():
            db.fix_sqlite()
            db.db.create_all()


class TestDB(_DBTestCase):
    def test_create_all(self):
        pass

    def test_get_or_create_get(self):
        with self.app.app_context(), db.db.session.begin():
            db.db.session.add(db.Component(name="nsname"))
            res = db.get_or_create(db.Component, name="nsname")
            self.assertEqual(res.name, "nsname")

    def test_get_or_create_create(self):
        with self.app.app_context(), db.db.session.begin():
            res = db.get_or_create(db.Component, name="nsname")
            self.assertIsNotNone(res.id)
            self.assertEqual(res.name, "nsname")


class TestGetByIdOrName(_DBTestCase):
    def setUp(self):
        super(TestGetByIdOrName, self).setUp()
        ctx = self.app.app_context()
        ctx.push()
        self.addCleanup(ctx.pop)
        self.component = db.Component(name="compname")
        db.db.session.add(self.component)
        db.db.session.flush()

    def test_by_id(self):
        res = db.Component.query.get_by_id_or_name(self.component.id)
        self.assertEqual(self.component, res)

    def test_by_name(self):
        res = db.Component.query.get_by_id_or_name(self.component.name)
        self.assertEqual(self.component, res)

    def test_by_id_fail(self):
        self.assertRaises(
            exceptions.NotFound,
            db.Component.query.get_by_id_or_name,
            self.component.id + 1,
        )

    def test_by_name_fail(self):
        self.assertRaises(
            exceptions.NotFound,
            db.Component.query.get_by_id_or_name,
            self.component.name + "_",
        )


class TestDBPrefixed(base.PrefixedTestCaseMixin, TestDB):
    pass


class TestEnvironmentHierarchyLevel(_DBTestCase):
    def setUp(self):
        super(TestEnvironmentHierarchyLevel, self).setUp()
        with self.app.app_context(), db.db.session.begin():
            session = db.db.session
            environment = db.Environment()
            session.add(environment)
            session.flush()
            self.environment_id = environment.id

    def _create_levels(self, num):
        session = db.db.session
        last_lvl = None
        for i in range(num):
            lvl = db.EnvironmentHierarchyLevel(
                environment_id=self.environment_id,
                parent=last_lvl,
                name="lvl%s" % (i,),
            )
            session.add(lvl)
            last_lvl = lvl
        session.flush()

    def _test_get_for_environment(self, num, expected):
        with self.app.app_context(), db.db.session.begin():
            self._create_levels(num)
            env = db.Environment(id=self.environment_id)
            res = db.EnvironmentHierarchyLevel.get_for_environment(env)
            level_names = [level.name for level in res]
        self.assertEqual(level_names, expected)

    def test_get_for_environment_empty(self):
        self._test_get_for_environment(0, [])

    def test_get_for_environment_one(self):
        self._test_get_for_environment(1, ['lvl0'])

    def test_get_for_environment_three(self):
        self._test_get_for_environment(3, ['lvl0', 'lvl1', 'lvl2'])


class TestEnvironmentHierarchyLevelPrefixed(base.PrefixedTestCaseMixin,
                                            TestEnvironmentHierarchyLevel):
    pass


class _RealDBTest(testscenarios.WithScenarios,
                  base.TestCase,
                  test_base.DbTestCase):
    scenarios = [
        ('sqlite', {'FIXTURE': test_base.DbFixture}),
        # ('mysql', {'FIXTURE': test_base.MySQLOpportunisticFixture}),
        ('postgres', {'FIXTURE': test_base.PostgreSQLOpportunisticFixture}),
    ]

    def get_alembic_config(self, engine):
        config = alembic_config.Config()
        config.set_main_option('sqlalchemy.url', str(engine.url))
        config.set_main_option(
            'script_location', tuning_box.get_migrations_dir())
        config.set_main_option('version_table', 'alembic_version')
        return config


class _RealDBPrefixedTest(base.PrefixedTestCaseMixin,
                          _RealDBTest):
    def get_alembic_config(self, engine):
        config = super(_RealDBPrefixedTest, self).get_alembic_config(
            engine)
        config.set_main_option('version_table', 'test_prefix_alembic_version')
        config.set_main_option('table_prefix', 'test_prefix_')
        return config


class TestMigrationsSync(_RealDBTest,
                         test_migrations.ModelsMigrationsSync):
    def get_metadata(self):
        return db.db.metadata

    def get_engine(self):
        return self.engine

    def db_sync(self, engine):
        config = self.get_alembic_config(engine)
        alembic_command.upgrade(config, 'head')


class TestMigrationsSyncPrefixed(_RealDBPrefixedTest,
                                 TestMigrationsSync):
    def include_object(self, object_, name, type_, reflected, compare_to):
        # ModelsMigrationsSync doesn't pass any config to MigrationContext
        # so alembic assumes 'alembic_version' table by default, not our
        # prefixed table
        if type_ == 'table' and name == 'test_prefix_alembic_version':
            return False

        return super(TestMigrationsSyncPrefixed, self).include_object(
            object_, name, type_, reflected, compare_to)


class TestRemoveFakeRootMigration(_RealDBTest):
    revision = '9ae15c85fa92'
    prefix = ''

    def setUp(self):
        super(TestRemoveFakeRootMigration, self).setUp()
        self.alembic_config = self.get_alembic_config(self.engine)
        script_dir = alembic_script.ScriptDirectory.from_config(
            self.alembic_config)
        self.migration_module = script_dir.get_revision(self.revision).module
        self.down_revision = self.migration_module.down_revision
        self.session = sa.orm.Session(bind=self.engine)
        self.addCleanup(self.session.close)

    def test_upgrade(self):
        alembic_command.upgrade(self.alembic_config, self.down_revision)
        abase = self.migration_module._get_autobase(self.prefix, self.engine)
        clss = abase.classes
        env = clss.Environment()
        env_level = clss.EnvironmentHierarchyLevel(environment=env, name='lvl')
        self.session.add(env_level)
        fake_root_1 = clss.EnvironmentHierarchyLevelValue(
            level_id=None, value=None, parent_id=None)
        self.session.add(fake_root_1)
        self.session.flush()
        child_1 = clss.EnvironmentHierarchyLevelValue(
            level_id=env_level.id, value="1", parent_id=fake_root_1.id)
        self.session.add(child_1)
        fake_root_2 = clss.EnvironmentHierarchyLevelValue(
            level_id=None, value=None, parent_id=None)
        self.session.add(fake_root_2)
        self.session.flush()
        child_2 = clss.EnvironmentHierarchyLevelValue(
            level_id=env_level.id, value="2", parent_id=fake_root_2.id)
        self.session.add(child_2)
        self.session.commit()
        alembic_command.upgrade(self.alembic_config, self.revision)
        ehlvs = self.session.query(clss.EnvironmentHierarchyLevelValue).all()
        for ehlv in ehlvs:
            self.assertIsNotNone(ehlv.level_id)
            self.assertIsNone(ehlv.parent_id)

    def test_downgrade(self):
        alembic_command.upgrade(self.alembic_config, self.revision)
        abase = self.migration_module._get_autobase(self.prefix, self.engine)
        clss = abase.classes
        env = clss.Environment()
        env_level = clss.EnvironmentHierarchyLevel(environment=env, name='lvl')
        self.session.add(env_level)
        self.session.flush()
        child_1 = clss.EnvironmentHierarchyLevelValue(
            level_id=env_level.id, value="1", parent_id=None)
        self.session.add(child_1)
        child_2 = clss.EnvironmentHierarchyLevelValue(
            level_id=env_level.id, value="2", parent_id=None)
        self.session.add(child_2)
        self.session.commit()
        alembic_command.downgrade(self.alembic_config, self.down_revision)
        fake_root = self.session.query(clss.EnvironmentHierarchyLevelValue) \
            .filter_by(parent_id=None) \
            .one()
        self.assertIsNone(fake_root.level_id)
        self.assertIsNone(fake_root.value)
        children = self.session.query(clss.EnvironmentHierarchyLevelValue) \
            .filter_by(parent_id=fake_root.id) \
            .all()
        self.assertItemsEqual(["1", "2"], [c.value for c in children])
        for child in children:
            self.assertEqual(fake_root.id, child.parent_id)
            self.assertEqual(env_level.id, child.level_id)


class TestRemoveFakeRootMigrationPrefixed(_RealDBPrefixedTest,
                                          TestRemoveFakeRootMigration):
    prefix = 'test_prefix_'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

from cliff import command
import yaml


def level_converter(value):
    levels = []
    for part in value.split(','):
        spl = part.split("=", 1)
        if len(spl) != 2:
            raise TypeError("Levels list should be in format "
                            "level1=value1,level2=value2")
        levels.append(tuple(spl))
    return levels

try:
    text_type = unicode
except NameError:
    text_type = str


def format_output(output, format_):
    if format_ == 'plain':
        if output is None:
            return ''
        if isinstance(output, text_type):
            if text_type is str:
                return output
            else:
                return output.encode('utf-8')
        format_ = 'json'
        # numbers, booleans, lists and dicts will be represented as JSON
    if format_ == 'json':
        return json.dumps(output)
    if format_ == 'yaml':
        # Usage of safe_dump here is crucial since PyYAML emits
        # "!!python/unicode" objects from unicode strings by defaul
        return yaml.safe_dump(output, default_flow_style=False)
    raise RuntimeError("Unknown format '{}'".format(format_))


class BaseCommand(command.Command):
    def get_client(self):
        return self.app.client


class FormattedCommand(BaseCommand):
    format_choices = ('json', 'yaml', 'plain')

    def get_parser(self, *args, **kwargs):
        parser = super(FormattedCommand, self).get_parser(*args, **kwargs)
        parser.add_argument(
            '--format',
            choices=self.format_choices,
            default='json',
            help="Desired format for return value",
        )
        return parser

    def run(self, parsed_args):
        res = self.take_action(parsed_args)
        self.app.stdout.write(format_output(res, parsed_args.format))
        return 0


class ResourceCommand(BaseCommand):
    def get_parser(self, *args, **kwargs):
        parser = super(ResourceCommand, self).get_parser(*args, **kwargs)
        parser.add_argument(
            '--env',
            type=int,
            required=True,
            help="ID of environment to get data from",
        )
        parser.add_argument(
            '--level',
            type=level_converter,
            default=[],
            help=("Level to get data from. Should be in format "
                  "parent_level=parent1,level=value2"),
        )
        parser.add_argument(
            '--resource',
            type=str,
            required=True,
            help="Name or ID of resource to get data from",
        )
        return parser

    def get_resource_url(self, parsed_args, last_part='values'):
        return '/environments/{}/{}resources/{}/{}'.format(
            parsed_args.env,
            ''.join('{}/{}/'.format(*e) for e in parsed_args.level),
            parsed_args.resource,
            last_part,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from werkzeug import routing

from tuning_box import converters
from tuning_box.tests import base


class TestLevels(base.TestCase):
    def setUp(self):
        super(TestLevels, self).setUp()
        self.map = routing.Map([
            routing.Rule('/smth/<levels:levels>values', endpoint='l'),
        ], converters={'levels': converters.Levels})
        self.mapad = self.map.bind('example.org', '/')

    def test_empty(self):
        route, kwargs = self.mapad.match('/smth/values')
        self.assertEqual(kwargs['levels'], [])

    def test_one(self):
        route, kwargs = self.mapad.match('/smth/level1/value1/values')
        self.assertEqual(kwargs['levels'], [('level1', 'value1')])

    def test_multi(self):
        route, kwargs = self.mapad.match(
            '/smth/level1/value1/level2/value2/values')
        self.assertEqual(kwargs['levels'],
                         [('level1', 'value1'), ('level2', 'value2')])

    def test_reverse_empty(self):
        res = self.mapad.build('l', {'levels': []})
        self.assertEqual(res, '/smth/values')

    def test_reverse_one(self):
        res = self.mapad.build('l', {'levels': [('level1', 'value1')]})
        self.assertEqual(res, '/smth/level1/value1/values')

    def test_reverse_multi(self):
        res = self.mapad.build(
            'l', {'levels': [('level1', 'value1'), ('level2', 'value2')]})
        self.assertEqual(res, '/smth/level1/value1/level2/value2/values')


class TestIdOrName(base.TestCase):
    def setUp(self):
        super(TestIdOrName, self).setUp()
        self.map = routing.Map([
            routing.Rule('/<id_or_name:id_or_name>', endpoint='i'),
        ], converters={'id_or_name': converters.IdOrName})
        self.mapad = self.map.bind('example.org', '/')

    def test_int(self):
        route, kwargs = self.mapad.match('/1')
        self.assertEqual(kwargs['id_or_name'], 1)

    def test_name(self):
        route, kwargs = self.mapad.match('/name')
        self.assertEqual(kwargs['id_or_name'], 'name')

    def test_name_with_slashes(self):
        route, kwargs = self.mapad.match('/name/with/slashes')
        self.assertEqual(kwargs['id_or_name'], 'name/with/slashes')

    def test_reverse_int(self):
        res = self.mapad.build('i', {'id_or_name': 1})
        self.assertEqual(res, '/1')

    def test_reverse_name(self):
        res = self.mapad.build('i', {'id_or_name': 'name'})
        self.assertEqual(res, '/name')

    def test_reverse_name_with_slashes(self):
        res = self.mapad.build('i', {'id_or_name': 'name/with/slashes'})
        self.assertEqual(res, '/name/with/slashes')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Add overrides to resource_values

Revision ID: ad192a40fd68
Revises: 967a44dd16d5
Create Date: 2016-03-25 21:26:19.170101

"""

# revision identifiers, used by Alembic.
revision = 'ad192a40fd68'
down_revision = '967a44dd16d5'
branch_labels = None
depends_on = None

from alembic import context
from alembic import op
import sqlalchemy as sa

import tuning_box.db


def upgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    op.add_column(table_prefix + 'resource_values', sa.Column(
        'overrides',
        tuning_box.db.Json(),
        server_default='{}',
        nullable=True,
    ))


def downgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    op.drop_column(table_prefix + 'resource_values', 'overrides')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging.config

from alembic import context
import sqlalchemy

from tuning_box import db

config = context.config
if config.get_main_option('table_prefix') is None:
    config.set_main_option('table_prefix', '')
if config.config_file_name:
    logging.config.fileConfig(config.config_file_name)
target_metadata = db.db.metadata


def run_migrations_offline():
    """Run migrations in 'offline' mode.

    This configures the context with just a URL
    and not an Engine, though an Engine is acceptable
    here as well.  By skipping the Engine creation
    we don't even need a DBAPI to be available.

    Calls to context.execute() here emit the given string to the
    script output.

    """
    context.configure(
        url=config.get_main_option('sqlalchemy.url'),
        version_table=config.get_main_option('version_table'),
        literal_binds=True,
    )

    with context.begin_transaction():
        context.run_migrations()


def run_migrations_online():
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    connectable = sqlalchemy.engine_from_config(
        config.get_section(config.config_ini_section),
        prefix='sqlalchemy.',
        poolclass=sqlalchemy.pool.NullPool,
    )

    with connectable.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            version_table=config.get_main_option('version_table'),
        )

        with context.begin_transaction():
            context.run_migrations()

if context.is_offline_mode():
    run_migrations_offline()
else:
    run_migrations_online()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import json

import yaml

from tuning_box.cli import base


class Set(base.ResourceCommand):
    url_last_part = 'values'

    def get_parser(self, *args, **kwargs):
        parser = super(Set, self).get_parser(*args, **kwargs)
        parser.add_argument(
            '--key',
            type=str,
            help="Name of key to set in the resource",
        )
        parser.add_argument(
            '--value',
            type=str,
            help="Value for a key to set in the resource",
        )
        parser.add_argument(
            '--type',
            choices=('null', 'int', 'str', 'json', 'yaml', 'bool'),
            help="Tyep of value passed in --value",
        )
        parser.add_argument(
            '--format',
            choices=('json', 'yaml'),
            help="Format of data passed to stdin",
        )
        return parser

    def verify_arguments(self, parsed_args):
        if parsed_args.key is None:  # no key
            if parsed_args.value is not None or parsed_args.type is not None:
                raise Exception("--value and --type arguments make sense only "
                                "with --key argument.")
            if parsed_args.format is None:
                raise Exception("Please specify format of data passed to stdin"
                                " to replace whole resource data.")
        elif parsed_args.value is not None:  # have key and value
            if parsed_args.format is not None:
                raise Exception("You shouldn't specify --format if you pass "
                                "value in command line, specify --type "
                                "instead.")
            if parsed_args.type == 'null':
                raise Exception("You shouldn't specify a value for 'null' type"
                                " because there can be only one.")
            if parsed_args.type is None:
                raise Exception("Please specify type of value passed in "
                                "--value argument to properly represent it"
                                " in the storage.")
        elif parsed_args.type != 'null':  # have key but no value
            if parsed_args.type is not None:
                raise Exception("--type specifies type for value provided in "
                                "--value but there is not --value argument")
            if parsed_args.format is None:
                raise Exception("Please specify format of data passed to stdin"
                                " to replace the key.")

    def get_value_to_set(self, parsed_args):
        type_ = parsed_args.type
        if type_ == 'null':
            return None
        elif type_ == 'bool':
            if parsed_args.value.lower() in ('1', 'true'):
                return True
            elif parsed_args.value.lower() in ('0', 'false'):
                return False
            else:
                raise Exception(
                    "Bad value for 'bool' type: '{}'. Should be one of '0', "
                    "'1', 'false', 'true'.".format(parsed_args.value))
        elif type_ == 'int':
            return int(parsed_args.value)
        elif type_ == 'str':
            return parsed_args.value
        elif type_ == 'json':
            return json.loads(parsed_args.value)
        elif type_ == 'yaml':
            return yaml.safe_load(parsed_args.value)
        elif type_ is None:
            if parsed_args.format == 'json':
                return json.load(self.app.stdin)
            elif parsed_args.format == 'yaml':
                docs_gen = yaml.safe_load_all(self.app.stdin)
                doc = next(docs_gen)
                guard = object()
                if next(docs_gen, guard) is not guard:
                    self.app.stderr.write("Warning: will use only first "
                                          "document from YAML stream")
                return doc
        assert False, "Shouldn't get here"

    def take_action(self, parsed_args):
        self.verify_arguments(parsed_args)
        value = self.get_value_to_set(parsed_args)

        client = self.get_client()
        resource_url = self.get_resource_url(parsed_args, self.url_last_part)
        if parsed_args.key is not None:
            resource = client.get(resource_url)
            resource[parsed_args.key] = value
        else:
            resource = value
        client.put(resource_url, resource)


class Override(Set):
    url_last_part = 'overrides'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import io

from requests_mock.contrib import fixture as req_fixture
import testscenarios

from tuning_box import cli
from tuning_box.cli import base as cli_base
from tuning_box import client as tb_client
from tuning_box.tests import base


class TestLevelsConverter(testscenarios.WithScenarios, base.TestCase):
    scenarios = [
        (s[0], dict(zip(('input', 'expected_result', 'expected_error'), s[1])))
        for s in [
            ('empty', ('', None, TypeError)),
            ('one', ('lvl=val', [('lvl', 'val')])),
            ('two', ('lvl1=val1,lvl2=val2', [('lvl1', 'val1'),
                                             ('lvl2', 'val2')])),
            ('no_eq', ('val', None, TypeError)),
            ('no_eq2', ('lvl1=val2,val', None, TypeError)),
            ('two_eq', ('lvl1=foo=baz', [('lvl1', 'foo=baz')])),
        ]
    ]

    input = None
    expected_result = None
    expected_error = None

    def test_levels(self):
        if self.expected_error:
            self.assertRaises(
                self.expected_error, cli_base.level_converter, self.input)
        else:
            res = cli_base.level_converter(self.input)
            self.assertEqual(self.expected_result, res)


class FormatOutputTest(testscenarios.WithScenarios, base.TestCase):
    scenarios = [
        (s[0], dict(zip(('output', 'format_', 'expected_result'), s[1])))
        for s in [
            ('none,plain', (None, 'plain', '')),
            ('none,json', (None, 'json', 'null')),
            # TODO(yorik-sar): look into why YAML return those elipsis
            ('none,yaml', (None, 'yaml', 'null\n...\n')),
            ('str,plain', (u"a string", 'plain', 'a string')),
            ('str,json', (u"a string", 'json', '"a string"')),
            ('str,yaml', (u"a string", 'yaml', 'a string\n...\n')),
            ('int,plain', (42, 'plain', '42')),
            ('int,json', (42, 'json', '42')),
            ('int,yaml', (42, 'yaml', '42\n...\n')),
            ('float,plain', (1.2, 'plain', '1.2')),
            ('float,json', (1.2, 'json', '1.2')),
            ('float,yaml', (1.2, 'yaml', '1.2\n...\n')),
            ('list,plain', ([1, 2], 'plain', '[1, 2]')),
            ('list,json', ([1, 2], 'json', '[1, 2]')),
            ('list,yaml', ([1, 2], 'yaml', '- 1\n- 2\n')),
            ('dict,plain', ({'a': 1}, 'plain', '{"a": 1}')),
            ('dict,json', ({'a': 1}, 'json', '{"a": 1}')),
            ('dict,yaml', ({'a': 1}, 'yaml', 'a: 1\n')),
        ]
    ]

    output = None
    format_ = None
    expected_result = None

    def test_format_output(self):
        res = cli_base.format_output(self.output, self.format_)
        self.assertEqual(self.expected_result, res)


class SafeTuningBoxApp(cli.TuningBoxApp):
    def __init__(self, client):
        super(SafeTuningBoxApp, self).__init__(
            client=client,
            **self.get_std_streams()
        )

    @staticmethod
    def get_std_streams():
        if bytes is str:
            io_cls = io.BytesIO
        else:
            io_cls = io.StringIO
        return {k: io_cls() for k in ('stdin', 'stdout', 'stderr')}

    def build_option_parser(self, description, version, argparse_kwargs=None):
        parser = super(SafeTuningBoxApp, self).build_option_parser(
            description, version, argparse_kwargs)
        parser.set_defaults(debug=True)
        return parser

    def get_fuzzy_matches(self, cmd):
        # Turn off guessing, we need exact failures in tests
        return []

    def run(self, argv):
        try:
            exit_code = super(SafeTuningBoxApp, self).run(argv)
        except SystemExit as e:
            exit_code = e.code
        assert exit_code == 0


class _BaseCLITest(base.TestCase):
    BASE_URL = 'http://somehost/prefix'

    def setUp(self):
        super(_BaseCLITest, self).setUp()
        client = tb_client.HTTPClient(self.BASE_URL)
        self.req_mock = self.useFixture(req_fixture.Fixture())
        self.cli = SafeTuningBoxApp(client=client)


class TestApp(_BaseCLITest):
    def test_help(self):
        self.cli.run(["--help"])
        self.assertEqual('', self.cli.stderr.getvalue())
        self.assertNotIn('Could not', self.cli.stdout.getvalue())


class TestGet(testscenarios.WithScenarios, _BaseCLITest):
    scenarios = [
        (s[0], dict(zip(('mock_url', 'args', 'expected_result'), s[1])))
        for s in [
            ('global,json', (
                '/environments/1/resources/1/values?effective',
                'get --env 1 --resource 1',
                '{"hello": "world"}',
            )),
            ('lowlevel,json', (
                '/environments/1/lvl1/value1/resources/1/values?effective',
                'get --env 1 --level lvl1=value1 --resource 1',
                '{"hello": "world"}',
            )),
            ('global,yaml', (
                '/environments/1/resources/1/values?effective',
                'get --env 1 --resource 1 --format yaml',
                'hello: world\n',
            )),
            ('lowlevel,yaml', (
                '/environments/1/lvl1/value1/resources/1/values?effective',
                'get --env 1 --level lvl1=value1 --resource 1 --format yaml',
                'hello: world\n',
            )),
            ('key,plain', (
                '/environments/1/resources/1/values?effective',
                'get --env 1 --resource 1 --key hello --format plain',
                'world',
            )),
            ('key,json', (
                '/environments/1/resources/1/values?effective',
                'get --env 1 --resource 1 --key hello --format json',
                '{"hello": "world"}',
            )),
            ('key,yaml', (
                '/environments/1/resources/1/values?effective',
                'get --env 1 --resource 1 --key hello --format yaml',
                'hello: world\n',
            )),
        ]
    ]

    mock_url = None
    args = None
    expected_result = None

    def test_get(self):
        self.req_mock.get(
            self.BASE_URL + self.mock_url,
            headers={'Content-Type': 'application/json'},
            json={'hello': 'world'},
        )
        self.cli.run(self.args.split())
        self.assertEqual(self.expected_result, self.cli.stdout.getvalue())


class TestSet(testscenarios.WithScenarios, _BaseCLITest):
    scenarios = [
        (s[0],
         dict(zip(('args', 'expected_body', 'should_get', 'stdin'), s[1])))
        for s in [
            ('json', ('--format json', {'a': 3}, False, '{"a": 3}')),
            ('yaml', ('--format yaml', {'a': 3}, False, 'a: 3')),
            ('key,json', ('--key b --format json', {'a': 1, 'b': {'a': 3}},
                          True, '{"a": 3}')),
            ('key,yaml', ('--key b --format yaml', {'a': 1, 'b': {'a': 3}},
                          True, 'a: 3')),
            ('key,null', ('--key b --type null', {'a': 1, 'b': None})),
            ('key,str', ('--key b --type str --value 4', {'a': 1, 'b': '4'})),
        ]
    ]

    args = None
    expected_body = None
    should_get = True
    stdin = None

    url_last_part = 'values'
    cmd = 'set'

    def test_set(self):
        url = self.BASE_URL + '/environments/1/lvl1/value1/resources/1/' + \
            self.url_last_part
        self.req_mock.put(url)
        if self.should_get:
            self.req_mock.get(
                url,
                headers={'Content-Type': 'application/json'},
                json={'a': 1, 'b': True},
            )
        args = [self.cmd] + ("--env 1 --level lvl1=value1 --resource 1 " +
                             self.args).split()
        if self.stdin:
            self.cli.stdin.write(self.stdin)
            self.cli.stdin.seek(0)
        self.cli.run(args)
        req_history = self.req_mock.request_history
        if self.should_get:
            self.assertEqual('GET', req_history[0].method)
        self.assertEqual('PUT', req_history[-1].method)
        self.assertEqual(self.expected_body, req_history[-1].json())


class TestOverride(TestSet):
    url_last_part = 'overrides'
    cmd = 'override'



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Initial revision

Revision ID: f16eb4eff7c
Revises:
Create Date: 2016-03-02 17:10:04.750584

"""

# revision identifiers, used by Alembic.
revision = 'f16eb4eff7c'
down_revision = None
branch_labels = None
depends_on = None

from alembic import context
from alembic import op
import sqlalchemy as sa

import tuning_box.db


def upgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    op.create_table(
        table_prefix + 'component',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('name', sa.String(length=128), nullable=True),
    )
    op.create_table(
        table_prefix + 'environment',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
    )
    op.create_table(
        table_prefix + 'namespace',
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('name', sa.String(length=128), nullable=True),
    )
    table_name = table_prefix + 'environment_components'
    op.create_table(
        table_name,
        sa.Column('environment_id', sa.Integer(), nullable=True),
        sa.Column('component_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['component_id'], [table_prefix + 'component.id'],
            name=table_name + '_component_id_fkey',
        ),
        sa.ForeignKeyConstraint(
            ['environment_id'], [table_prefix + 'environment.id'],
            name=table_name + '_environment_id_fkey',
        ),
    )
    table_name = table_prefix + 'environment_hierarchy_level'
    op.create_table(
        table_name,
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('environment_id', sa.Integer(), nullable=True),
        sa.Column('name', sa.String(length=128), nullable=True),
        sa.Column('parent_id', sa.Integer(), nullable=True),
        sa.ForeignKeyConstraint(
            ['environment_id'], [table_prefix + 'environment.id'],
            name=table_name + '_environment_id_fkey',
        ),
        sa.ForeignKeyConstraint(
            ['parent_id'], [table_prefix + 'environment_hierarchy_level.id'],
            name=table_name + '_parent_id_fkey',
        ),
        sa.UniqueConstraint(
            'environment_id', 'name',
            name=table_name + '_environment_id_name_key',
        ),
        sa.UniqueConstraint(
            'environment_id', 'parent_id',
            name=table_name[:-4] + '_environment_id_parent_id_key',
        ),
    )
    table_name = table_prefix + 'schema'
    op.create_table(
        table_name,
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('name', sa.String(length=128), nullable=True),
        sa.Column('component_id', sa.Integer(), nullable=True),
        sa.Column('namespace_id', sa.Integer(), nullable=True),
        sa.Column('content', tuning_box.db.Json(), nullable=True),
        sa.ForeignKeyConstraint(
            ['component_id'], [table_prefix + 'component.id'],
            name=table_name + '_component_id_fkey'),
        sa.ForeignKeyConstraint(
            ['namespace_id'], [table_prefix + 'namespace.id'],
            name=table_name + '_namespace_id_fkey'),
    )
    table_name = table_prefix + 'template'
    op.create_table(
        table_name,
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('name', sa.String(length=128), nullable=True),
        sa.Column('component_id', sa.Integer(), nullable=True),
        sa.Column('content', tuning_box.db.Json(), nullable=True),
        sa.ForeignKeyConstraint(
            ['component_id'], [table_prefix + 'component.id'],
            name=table_name + '_component_id_fkey',
        ),
    )
    table_name = table_prefix + 'environment_hierarchy_level_value'
    op.create_table(
        table_name,
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('level_id', sa.Integer(), nullable=True),
        sa.Column('parent_id', sa.Integer(), nullable=True),
        sa.Column('value', sa.String(length=128), nullable=True),
        sa.ForeignKeyConstraint(
            ['level_id'], [table_prefix + 'environment_hierarchy_level.id'],
            name=table_name + '_level_id_fkey',
        ),
        sa.ForeignKeyConstraint(
            ['parent_id'], [table_name + '.id'],
            name=table_name + '_parent_id_fkey',
        ),
    )
    table_name = table_prefix + 'environment_schema_values'
    op.create_table(
        table_name,
        sa.Column('id', sa.Integer(), nullable=False, primary_key=True),
        sa.Column('environment_id', sa.Integer(), nullable=True),
        sa.Column('schema_id', sa.Integer(), nullable=True),
        sa.Column('level_value_id', sa.Integer(), nullable=True),
        sa.Column('values', tuning_box.db.Json(), nullable=True),
        sa.ForeignKeyConstraint(
            ['environment_id'], [table_prefix + 'environment.id'],
            name=table_name + '_environment_id_fkey',
        ),
        sa.ForeignKeyConstraint(
            ['level_value_id'],
            [table_prefix + 'environment_hierarchy_level_value.id'],
            name=table_name + '_level_value_id_fkey',
        ),
        sa.ForeignKeyConstraint(
            ['schema_id'], [table_prefix + 'schema.id'],
            name=table_name + '_schema_id_fkey',
        ),
        sa.UniqueConstraint(
            'environment_id', 'schema_id', 'level_value_id',
            name=table_name[:-6] + 'environment_id_schema_id_leve_key',
        ),
    )


def downgrade():
    table_prefix = context.config.get_main_option('table_prefix')
    op.drop_table(table_prefix + 'environment_schema_values')
    op.drop_table(table_prefix + 'environment_hierarchy_level_value')
    op.drop_table(table_prefix + 'template')
    op.drop_table(table_prefix + 'schema')
    op.drop_table(table_prefix + 'environment_hierarchy_level')
    op.drop_table(table_prefix + 'environment_components')
    op.drop_table(table_prefix + 'namespace')
    op.drop_table(table_prefix + 'environment')
    op.drop_table(table_prefix + 'component')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from cliff import app
from cliff import commandmanager

import tuning_box


class TuningBoxApp(app.App):
    def __init__(self, client, **kwargs):
        super(TuningBoxApp, self).__init__(
            description='Tuning Box - configuration storage for your cloud',
            version=tuning_box.__version__,
            command_manager=commandmanager.CommandManager('tuning_box.cli'),
            **kwargs
        )
        self.client = client



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

from werkzeug import routing
from werkzeug import urls


class Levels(routing.BaseConverter):
    """Converter that maps nested levels to list of tuples.

    For example, "level1/value1/level2/value2/" is mapped to
    [("level1", "value1"), ("level2", "value2")].

    Note that since it can be empty it includes following "/":

        Rule('/smth/<levels:levels>values')

    will parse "/smth/values" and "/smth/level1/value1/values".
    """

    regex = "([^/]+/[^/]+/)*"

    def to_python(self, value):
        spl = value.split('/')
        return list(zip(spl[::2], spl[1::2]))

    def to_url(self, value):
        parts = itertools.chain.from_iterable(value)
        quoted_parts = (urls.url_quote(p, charset=self.map.charset, safe='')
                        for p in parts)
        return ''.join(p + '/' for p in quoted_parts)


class IdOrName(routing.BaseConverter):
    """Converter that matches either int or URL part including "/" as string"""

    regex = '[^/].*?'

    def to_python(self, value):
        try:
            return int(value)
        except ValueError:
            return value

    def to_url(self, value):
        return super(IdOrName, self).to_url(str(value))

ALL = {
    'levels': Levels,
    'id_or_name': IdOrName,
}
